from group_selfies import GroupGrammar
from rdkit.Chem.rdmolfiles import MolFromSmiles
from molvs import standardize_smiles
import selfies as sf
import safe
from deepsmiles import Converter
import pandas as pd
from datasets import load_dataset, Dataset, DatasetDict
from pandarallel import pandarallel
import argparse
import os

# Initialize pandarallel with a progress bar and 32 workers
pandarallel.initialize(progress_bar=True, nb_workers=32)


def convert_smiles_to_representations(smiles: str):
    """Convert SMILES to various representations (standard, DeepSMILES, SELFIES, SAFE, Group SELFIES).

    :param smiles: SMILES string to convert
    :return: Dictionary with various representations of the SMILES
    """
    try:
        # Standardize the SMILES string
        standardized_smiles = standardize_smiles(smiles)
        molecule = MolFromSmiles(standardized_smiles)
    except Exception as e:
        print(f"Error standardizing SMILES '{smiles}': {e}")
        standardized_smiles = None

    try:
        # Convert to DeepSMILES
        deep_smiles_converter = Converter(rings=True, branches=True)
        deep_smiles = deep_smiles_converter.encode(standardized_smiles)
    except Exception as e:
        print(f"Error converting SMILES to DeepSMILES '{standardized_smiles}': {e}")
        deep_smiles = None

    try:
        # Convert to SELFIES
        selfies_representation = sf.encoder(standardized_smiles)
    except Exception as e:
        print(f"Error converting SMILES to SELFIES '{standardized_smiles}': {e}")
        selfies_representation = None

    try:
        # Convert to SAFE
        safe_representation = safe.encode(standardized_smiles)
    except Exception as e:
        print(f"Error converting SMILES to SAFE '{standardized_smiles}': {e}")
        safe_representation = None

    try:
        # Convert to Group SELFIES
        grammar = GroupGrammar(vocab={})  # Empty grammar
        extracted_groups = grammar.extract_groups(molecule)
        group_selfies_representation = grammar.encoder(molecule, extracted_groups)
    except Exception as e:
        print(f"Error converting SMILES to Group SELFIES '{standardized_smiles}': {e}")
        group_selfies_representation = None

    # Return all representations as a dictionary
    return {
        "Standard SMILES": standardized_smiles,
        "Deep SMILES": deep_smiles,
        "SELFIES": selfies_representation,
        "SAFE": safe_representation,
        "Group SELFIES": group_selfies_representation,
    }


def process_and_save_smiles_dataframe(smiles_df: pd.DataFrame, output_path: str):
    """Process SMILES in the DataFrame and save the results.

    :param smiles_df: DataFrame containing SMILES strings
    :param output_path: Path to save the processed DataFrame
    """
    # Apply convert_smiles_to_representations in parallel and convert to a DataFrame
    processed_df = (
        smiles_df["SMILES"]
        .parallel_apply(convert_smiles_to_representations)
        .apply(pd.Series)
    )
    processed_df.dropna(inplace=True)  # Drop rows with any missing values
    processed_df.drop_duplicates(inplace=True)  # Drop duplicate rows
    processed_df = processed_df.rename(
        columns={"Standard SMILES": "SMILES"}
    )  # Rename column
    processed_df.to_parquet(output_path)  # Save to parquet file


def process_dataset(dataset_name: str):
    """Process dataset from Hugging Face or local directory.

    :param dataset_name: Name of the dataset (Hugging Face repo or local directory)
    """
    try:
        # Attempt to load the dataset from Hugging Face
        hf_data = load_dataset(
            f"MolGen/{dataset_name}", token="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"
        )
        print(f"Loaded dataset from Hugging Face: {dataset_name}")

        # Process each split in the dataset
        for split in list(hf_data.keys()):
            smiles_df = pd.DataFrame(hf_data[split])  # Convert to DataFrame
            output_path = f"./processed/{dataset_name}_{split}.parquet"
            process_and_save_smiles_dataframe(
                smiles_df, output_path
            )  # Process and save
            # Reload the processed data and update the dataset
            hf_data[split] = Dataset.from_pandas(
                pd.read_parquet(output_path), preserve_index=False
            )
            print(
                f"Total number of rows removed in {split}: {len(smiles_df) - len(hf_data[split])}"
            )

        # Push the processed dataset back to Hugging Face Hub
        hf_data.push_to_hub(
            f"MolGen/{dataset_name}",
            private=True,
            token="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",
        )
    except Exception as e:
        print(f"Could not load dataset from Hugging Face: {e}")

        # Attempt to load the dataset from the local directory
        raw_dir = f"./raw/{dataset_name}"
        processed_dir = f"./processed/{dataset_name}"

        # Check if the raw directory exists
        if not os.path.exists(raw_dir):
            raise FileNotFoundError(
                f"Dataset not found in Hugging Face or local directory: {dataset_name}"
            )

        os.makedirs(processed_dir, exist_ok=True)  # Ensure processed directory exists
        raw_files = os.listdir(raw_dir)  # List files in the raw directory
        print(f"Found {len(raw_files)} files in local directory.")

        processed_files = os.listdir(
            processed_dir
        )  # List files in the processed directory
        # Exclude already processed files from the list
        files_to_process = [file for file in raw_files if file not in processed_files]
        print(f"Processing {len(files_to_process)} new files.")

        # Process each file
        for file_name in files_to_process:
            smiles_df = pd.read_parquet(
                os.path.join(raw_dir, file_name)
            )  # Read the parquet file
            output_path = os.path.join(processed_dir, file_name)  # Define output path
            process_and_save_smiles_dataframe(
                smiles_df, output_path
            )  # Process and save


if __name__ == "__main__":
    # Parse command-line arguments
    parser = argparse.ArgumentParser(
        description="Process SMILES files in a folder or Hugging Face dataset."
    )
    parser.add_argument(
        "--dataset",
        help="The dataset name (Hugging Face repo or local)",
        default="pubchem",
    )
    args = parser.parse_args()

    # Process the dataset
    process_dataset(args.dataset)
