from molvs import standardize_smiles
from rdkit.Chem.rdmolfiles import MolFromSmiles
import selfies
import pandas as pd
from pandarallel import pandarallel
import argparse
import os
import pyarrow as pa
from concurrent.futures import ThreadPoolExecutor

# Initialize pandarallel with a progress bar
pandarallel.initialize(progress_bar=True)

def std_smiles_with_timeout(smiles: str) -> str | None:
    """Standardizes a SMILES string with a timeout.
    
    :param smiles: SMILES string
    :return: Standardized SMILES string or None if an error occurs
    """
    try:
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(std_smiles, smiles)
            return future.result(timeout=10)  # 10 seconds timeout
    except Exception as e:  # Catching all exceptions now
        print(f"Error processing SMILES {smiles}: {e}")  # Optional: Log the error for debugging
        return None

def smiles_to_selfies_with_timeout(smiles: str) -> str | None:
    """Converts a SMILES string to a SELFIES string with a timeout.
    
    :param smiles: SMILES string
    :return: SELFIES string or None if an error occurs
    """
    try:
        with ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(smiles_to_selfies, smiles)
            return future.result(timeout=10)  # 10 seconds timeout
    except Exception as e:  # Catching all exceptions now
        print(f"Error converting to SELFIES {smiles}: {e}")  # Optional: Log the error for debugging
        return None

def smiles_to_selfies(smiles: str) -> str | None:
    """Converts SMILES to SELFIES.
    
    :param smiles: SMILES string
    :return: SELFIES string or None if an error occurs
    """
    try:
        # Convert canonical SMILES to SELFIES
        selfies_string = selfies.encoder(smiles)
        return selfies_string
    except BaseException:
        return None

def std_smiles(smiles: str) -> str | None:
    """Standardizes a SMILES string.
    
    :param smiles: SMILES string
    :return: Standardized SMILES string or None if an error occurs
    """
    try:
        standard = standardize_smiles(smiles)
        MolFromSmiles(standard)  # Ensure the standardized SMILES is valid
        return standard
    except BaseException:
        return None

def process_file(file_path, smiles_column_name, output_directory):
    """Processes a file containing SMILES strings, standardizes them, and converts them to SELFIES.
    
    :param file_path: Path to the input file
    :param smiles_column_name: Column name containing SMILES strings
    :param output_directory: Directory to save processed files
    """
    print(file_path)
    file_extension = file_path.split(".")[-1]
    
    # Read the input file based on its extension
    if file_extension in ["csv", "txt", "smiles"]:
        df = pd.read_csv(file_path)
    elif file_extension == "smi":
        df = pd.read_csv(file_path, sep='\t', header=None, names=['SMILES', 'ZINC ID'])
    elif file_extension == "parquet":
        df = pd.read_parquet(file_path)
    elif file_extension in ["xlsx", "xls"]:
        df = pd.read_excel(file_path)
    elif file_extension == "arrow":
        with pa.memory_map(file_path, 'r') as source:
            stream = pa.ipc.open_stream(source)
            table = stream.read_all()
        df = table.to_pandas()
    else:
        raise ValueError(f"Unsupported file type: {file_extension}")

    # Define output file path
    output_file_path = os.path.join(output_directory, os.path.basename(file_path).replace(file_extension, "parquet"))

    # Drop rows with missing values
    df = df.dropna()

    # Standardize SMILES strings with a timeout
    df["STANDARDISED SMILES"] = df[smiles_column_name].parallel_apply(std_smiles_with_timeout)
    df = df.dropna()  # Drop rows with failed standardization

    # Convert standardized SMILES to SELFIES with a timeout
    df["SELFIES"] = df["STANDARDISED SMILES"].parallel_apply(smiles_to_selfies_with_timeout)
    df = df.dropna(subset=["STANDARDISED SMILES", "SELFIES"])  # Drop rows with failed conversions

    # Select and rename columns
    df = df[["STANDARDISED SMILES", "SELFIES"]]
    df = df.rename(columns={"STANDARDISED SMILES": "SMILES"})
    
    # Save processed data to a parquet file
    df.to_parquet(output_file_path)
    
    # Remove duplicate rows and save cleaned data
    df_cleaned = df.drop_duplicates(subset=["SMILES", "SELFIES"])
    modified_path = output_file_path.split(".parquet")[0] + "_cleaned" + ".parquet"
    df_cleaned.to_parquet(modified_path)
    print(f"Processed file saved to: {output_file_path}")

def main(folder_path, smiles_column_name, output_directory):
    """Main function to process all files in a folder.
    
    :param folder_path: Path to the folder containing input files
    :param smiles_column_name: Column name containing SMILES strings
    :param output_directory: Directory to save processed files
    """
    for file_name in os.listdir(folder_path):
        file_path = os.path.join(folder_path, file_name)
        if os.path.isfile(file_path):
            process_file(file_path, smiles_column_name, output_directory)

if __name__ == "__main__":
    # Argument parser for command line options
    parser = argparse.ArgumentParser(description="Process files in a folder.")
    parser.add_argument("--folder_path", help="The path of the input folder to process")
    parser.add_argument("--smiles_column_name", default="SMILES", help="The name of the column containing the SMILES data")
    parser.add_argument("--output_directory", default="./processed_files", help="The directory to save processed files")
    args = parser.parse_args()

    # Ensure the output directory exists
    os.makedirs(args.output_directory, exist_ok=True)
    
    # Process the input folder
    main(args.folder_path, args.smiles_column_name, args.output_directory)